(*
 * LANGUAGE    : ANS Forth 
 * PROJECT     : Forth Environments
 * DESCRIPTION : SOD64 virtual stack machine
 * CATEGORY    : Tool
 * AUTHOR      : L.C. Benschop, 1994
 * LAST CHANGE : Marcel Hendrix, Saturday, March 08, 2008, 10:34 AM.
 *)


	NEEDS -miscutil

	REVISION -cross  " SOD Cros-compiler   Version 0.01 "

	PRIVATES

DOC
(*
	CROSS COMPILER FOR SOD-64, THE STACK ORIENTED DESIGN PROCESSOR 
	Copyright 1994 L.C. Benschop Eindhoven, The Netherlands.
	The program is released under the GNU General Public License version 2.
	There is NO WARRANTY.

	This serves as an introduction to Forth cross compiling, so it is excessively 
	commented.  

	This cross compiler can be run on any ANS Forth with the necessary 
	extension wordset that is at least 32-bit, including SOD-64 Forth. 

	It creates the memory image of a new Forth system that is to be run 
	by the SOD-64 virtual processor. (or maybe a real SOD-64 processor if 
	it will ever be made.) 

	The cross compiler (or meta compiler or target compiler) is similar 
	to a regular Forth compiler, except that it builds definitions in 
	a dictionary in the memory image of a different Forth system. 
	We call this the target dictionary in the target space of the 
	target system.  

	As the new definitions are for a different Forth system, the cross 
	compiler cannot EXECUTE them. Neither can it easily find the new 
	definitions in the target dictionary. Hence a shadow definition 
	for each target definition is made in the normal Forth dictionary. 

	The names of the new definitions overlap with the names of existing
	elementary Forth words. Therefore they need to be in a wordlist 
	different from the normal Forth wordlist.
*)
ENDDOC

-- PART 1: THE VOCABULARIES.

VOCABULARY TARGET

DOC
(*
	This vocabulary will hold shadow definitions for all words that are in
	the target dictionary. When a shadow definition is executed, it 
	performs the compile action in the target dictionary.
*)
ENDDOC


VOCABULARY TRANSIENT

DOC
(*
	This vocabulary will hold definitions that must be executed by the
	host system ( the system on which the cross compiler runs) and that
	compile to the target system.

	Expl: The word IF occurs in all three vocabularies. The word IF in the
	      FORTH vocabulary is run by the host system and is used when
	      compiling host definitions. A different version is in the
	      TRANSIENT vocabulary. This one runs on the host system and
	      is used when compiling target definitions. The version in the
	      TARGET vocabulary is the version that will run on the target
	      system. 
*)
ENDDOC

\ : \D ; \ Uncomment one of these. If uncommented, display debug info.
: \D POSTPONE \ ; IMMEDIATE 

-- PART 2: THE TARGET DICTIONARY SPACE.

-- Next define the target space and the words to access it.

#200000 CONSTANT IMAGE_SIZE

CREATE IMAGE IMAGE_SIZE CHARS ALLOT	-- This space contains the target image.
       IMAGE IMAGE_SIZE 0 FILL      	-- Initialize it to zero.

-- Fetch and store characters in the target space.
: C@-T ( t-offs -- c ) CHARS IMAGE + C@ ;
: C!-T ( c t-offs -- ) CHARS IMAGE + C! ;

-- Fetch and store cells in the target space.
-- SOD64 is big endian 64 bit so store explicitly big-endian.
: @-T  ( t-offs -- x )
       CHARS IMAGE + D@ BSWAP SWAP BSWAP ;
: !-T  ( x t-offs -- )
       CHARS IMAGE + >R BSWAP SWAP BSWAP R> D! ;

-- A dictionary is constructed in the target space. Here are the primitives
-- to maintain the dictionary pointer and to reserve space.

VARIABLE DP-T				\ Dictionary pointer for target dictionary.
0 DP-T !				\ Initialize it to zero, SOD starts at 0.

: THERE ( -- t-addr ) DP-T @ ;		\ Equivalent of HERE in target space.                                
: ALLOT-T ( n -- ) DP-T +! ;		\ Reserve n bytes in the dictionary.
: CHARS-T ( n1 -- n2 ) ;      
: CELLS-T ( n1 -- n2 ) 3 LSHIFT ;	\ Cells are 8 chars.
: =TCELL  ( -- u ) 1 CELLS-T ;		\ SOD only accesses cells at aligned addresses.
: =TCHAR  ( -- u ) 1 ;

: ALIGN-T ( -- )  
	BEGIN  THERE 7 AND  
	WHILE  1 ALLOT-T  
	REPEAT ;

: ALIGNED-T ( n1 -- n2 ) 7 + -8 AND ; 
: C,-T  ( c -- )  THERE C!-T =TCHAR ALLOT-T ;
: ,-T   ( x -- )  THERE !-T  =TCELL ALLOT-T ;

: PLACE-T ( c-addr len t-addr --- ) 	\ Move counted string to target space.
  2DUP C!-T 1+ CHARS IMAGE + 
  SWAP CHARS CMOVE ;      

-- After the Forth system is constructed, its image must be saved.
: SAVE-IMAGE ( "name" --- )
  BL WORD COUNT W/O BIN CREATE-FILE ABORT" Can't open file" >R
  IMAGE THERE R@ WRITE-FILE ABORT" Can't write file" 
  R> CLOSE-FILE ABORT" Can't close file" ;

-- PART 3: CREATING NEW DEFINITIONS IN THE TARGET SYSTEM.

-- These words create new target definitions, both the shadow definition
-- and the header in the target dictionary. The layout of target headers
-- can be changed but FIND in the target system must be changed accordingly. 

VARIABLE TLINK  0 TLINK !		\ This variable points to the name
					\ of the last definition.

: "HEADER ( -- )
  >IN @ CREATE >IN !			\ Create the shadow definition.
  ALIGN-T TLINK @ U>D ,-T		\ Lay out the link field.   
  BL WORD                            
  COUNT DUP >R THERE PLACE-T		\ Place name in target dictionary. 
  THERE TLINK !               
  THERE C@-T #128 OR THERE C!-T  
  R> 1+ ALLOT-T ALIGN-T ; 		\ Set bit 7 of count byte as a marker.

: MACRO ( -- )
  TLINK @ DUP C@-T #32 OR SWAP C!-T ; 	\ Set the MACRO bit of last name. This indicates to the compiler in the
					\ target Forth system that this def may be expanded.

ALSO TRANSIENT DEFINITIONS 

: IMMEDIATE 
  TLINK @ DUP C@-T #64 OR SWAP C!-T ; 	\ Set the IMMEDIATE bit of last name.

PREVIOUS DEFINITIONS

-- PART 4: CODE GENERATION

DOC
(*
	SOD-64 Forth is a native code Forth for an unusual machine, the SOD64.
	Instructions are 64 bit cells. They are either calls to other definitions
	(if bit 0 and 1 are zero) or conditional jumps (if bit 1 is 1 and bit
	0 is 0, or packs of six 5-bit opcodes (bit 0 is 1). In that case bit 63
	indicates subroutine return. 

	Forth primitives such as + and R> are single opcodes. The compiler 
	compiles them as opcodes rather than calls to a definition. Other
	definitions such as - and ! consist of only a few opcodes and are
	expanded by the compiler into the constituent opcodes.
*)
ENDDOC

VARIABLE STATE-T   0 STATE-T !		\ State variable for cross compiler.

: T]	1 STATE-T ! ;
: T[	0 STATE-T ! ;

VARIABLE CSP				\ Stack pointer checking between : and ;

: !CSP	DEPTH CSP ! ;
: ?CSP	DEPTH CSP @ - 
	   IF  .S 
	       TRUE ABORT" ?CSP :: incomplete control structure" 
	ENDIF ;

$8000000000000000. 2CONSTANT RET-CODE 	\ Return bit.

VARIABLE OPLOC      0 OPLOC   !		\ Target address where opcodes are appended.
VARIABLE OPSHIFT  #63 OPSHIFT !		\ Number of bits to shift next opcode.

: CODEFLUSH	  #64 OPSHIFT ! ;	\ Indicate that no more opcodes 
					\ may be filled in at current location.
					\ The unused opcodes are left 0 (noop).

: INSERT-OPCODE ( i -- )		\ Insert opcode i (0..31) into definition.
	OPSHIFT @ #60 
	 > IF
		THERE OPLOC !  1. ,-T	\ Create a new 'twelvepack' instruction
		1 OPSHIFT !		\ where opcodes may be inserted.
	ENDIF

	U>D OPSHIFT @ DLSHIFT
	OPLOC @ @-T DOR  
	OPLOC @ !-T 			\ Or the shifted opcode.

	5 OPSHIFT +! ;			\ Next opcode must be shifted 5 more.

: LITERAL-T ( x -- ) 
	#31 INSERT-OPCODE  ,-T ;

TRANSIENT DEFINITIONS FORTH

-- Now define the words that compile code. 

: OPCODE ( c -- )
	"HEADER  DUP C, 1 LSHIFT U>D 
	RET-CODE 1. DOR DOR ,-T 
	MACRO				\ Create an executable
	DOES> C@ INSERT-OPCODE ;	\ target definition.
  

: : ( -- )
	!CSP "HEADER THERE , T] 
  	DOES> @ CODEFLUSH U>D ,-T ;

: M:  	"HEADER THERE , MACRO		\ M: makes a definition identical to that made by :
   	T]                     		\ but with macro bit set. If executed it copies 
					\ all opcodes from the macro definition to the 
					\ current definition. The macro itself is just
					\ a target definition that can be executed.
   	DOES> @ 
	  BEGIN
	    DUP @-T			\ Get next instruction.
	    OVER 1 AND 0= ABORT" Instruction not allowed in macro" 1 DRSHIFT
	    ROT =TCELL + -ROT		\ Increment source address.
	    #12 0 DO			\ Pick apart the twelve opcodes.
		       OVER #31 AND 
		         IF		\ Skip opcode 0 (noop)
				OVER #31 AND 
				DUP INSERT-OPCODE
				#31 = IF 
					\ Copy the literal for opcode 31 (lit) 
					2 PICK @-T ,-T  
					ROT =TCELL + -ROT
				   ENDIF
		      ENDIF
		      5 DRSHIFT    
		LOOP
	    DROP 4 AND 			\ Repeat until return bit encountered.  
          UNTIL DROP ;

: ; RET-CODE OPSHIFT @ #64 = 		\ If no current opcode location then 
	   IF 1. DOR ,-T          	\ Make new instruction with twelve noops and return.
	 ELSE	OPLOC @ @-T  DOR
	 	OPLOC @ !-T		\ else set return bit of current instruction.
	ENDIF  
	CODEFLUSH T[ ?CSP ;		\ Quit compilation state.

FORTH DEFINITIONS

-- PART 5: FORWARD REFERENCES 

-- Some definitions are referenced before they are defined. A definition
-- in the TRANSIENT voc is created for each forward referenced definition.
-- This links all addresses together where the forward reference is used.
-- The word RESOLVE stores the real address everywhere it is needed.    

: FORWARD  
	CREATE 	-1 ,			\ Store head of list in the definition.
	DOES> 	CODEFLUSH 		\ Reserve a cell in the dictionary
	 	DUP @ U>D ,-T 		\ where the call to the forward definition must come.
		THERE 1 CELLS-T - 	\ As the call address is unknown, store link to next 
		SWAP ! ;		\ reference instead.

: RESOLVE
  	ALSO TARGET   >IN  @ 		\ Find the resolving word in the 
	' >BODY @ >R  >IN  !		\ target vocabulary and take the CFA out of the definition.
  	TRANSIENT ' >BODY  @		\ Find the forward ref word in the
					\ TRANSIENT VOC and take list head.   
	BEGIN
	   DUP -1 <>			\ Traverse all the links until end.       
	WHILE
	   DUP @-T DROP	SWAP		\ Take address of next link from dict.
	   R@ U>D  ROT !-T 		\ Set resolved address in dict.
	REPEAT 
	DROP -R PREVIOUS ;


-- PART 6: DEFINING WORDS.

TRANSIENT DEFINITIONS FORTH
 
FORWARD DOVAR				\ Dovar is the runtime part of a variable.

: VARIABLE 
	"HEADER THERE ,			\ Create a variable.
	[ ALSO TRANSIENT ] DOVAR [ PREVIOUS ]  
		0. ,-T
	DOES> 	@ =TCELL + U>D
		LITERAL-T ;		\ Compile var address as a literal for speed.  

: CONSTANT ( x -- )
	"HEADER	THERE ,
  		RET-CODE 1. DOR 
		#31 2* U>D DOR ,-T	\ Assemble the instruction LIT with RETURN.
  		( x) ,-T
  	DOES> 	@ =TCELL + @-T		\ Compile const as a literal for speed.
		LITERAL-T ; 

FORTH DEFINITIONS

: T' ( -- t-offs ) 			\ Find the execution token of a target definition.
	ALSO TARGET ' >BODY @ 		\ Get the address from the shadow definition.
	PREVIOUS ;

: >BODY-T ( t-offs1 -- t-offs2 ) 	\ Convert executing token to param address.
	=TCELL + ;

-- PART 7: COMPILING WORDS 

TRANSIENT DEFINITIONS FORTH

-- The TRANSIENT definitions for IF, THEN etc. compile the conditional
-- branch instructions of SOD-64. This is the address with bit 1 set,
-- this is the address + 2. SOD has no unconditional branch. It is 
-- composed of opcode 28 (which pushes 0 onto the stack) and the
-- conditional branch. 

: BEGIN		CODEFLUSH THERE ; 			( -- t-offs )
: UNTIL 	CODEFLUSH 2+ U>D ,-T ; 			( t-offs -- )
: IF 		CODEFLUSH THERE  =TCELL ALLOT-T ;	( -- t-offs )
: THEN 		CODEFLUSH THERE 2+ U>D ROT !-T ;	( t-offs -- )
: ELSE 		#28 INSERT-OPCODE [ ALSO TRANSIENT ] IF SWAP THEN [ PREVIOUS ] ; 
: WHILE 	[ ALSO TRANSIENT ] IF [ PREVIOUS ] SWAP ; 
: REPEAT 	#28 INSERT-OPCODE [ ALSO TRANSIENT ] UNTIL THEN [ PREVIOUS ] ; 

FORWARD (DO)
FORWARD (LOOP)
FORWARD (.")
FORWARD (ABORT")
FORWARD (POSTPONE)

: DO 		[ ALSO TRANSIENT ] (DO)       [ PREVIOUS ] THERE ;	( -- t-offs )
: LOOP 		[ ALSO TRANSIENT ] (LOOP)     [ PREVIOUS ] U>D ,-T ;	( t-offs -- )
: ." 		[ ALSO TRANSIENT ] (.")       [ PREVIOUS ] 
		&" WORD COUNT DUP 1+ >R 
      		THERE PLACE-T 
		R> ALLOT-T ALIGN-T ;
: ABORT" 	[ ALSO TRANSIENT ] (ABORT")   [ PREVIOUS ] 
		&" WORD COUNT DUP 1+ >R 
      		THERE PLACE-T 
		R> ALLOT-T ALIGN-T ;
: POSTPONE 	[ ALSO TRANSIENT ] (POSTPONE) [ PREVIOUS ] 
		T' U>D ,-T ;

: [.S]		.S ; IMMEDIATE
: \		POSTPONE \ ; IMMEDIATE
: \G		POSTPONE \ ; IMMEDIATE
: ( 		POSTPONE ( ; IMMEDIATE 	-- Move duplicates of comment words to TRANSIENT
: CHARS-T  	CHARS-T ; 		-- Also words that must be executed while cross compiling.
: CELLS-T  	CELLS-T ;
: ALLOT-T  	ALLOT-T ;

FORTH DEFINITIONS

-- PART 8: THE CROSS COMPILER ITSELF.

VARIABLE DPL

: NUMBER? ( c-addr -- d flag )
	-1 DPL !
	BASE @ >R
	COUNT   
	OVER C@ '-' = DUP >R  IF     1 /STRING  ENDIF	\ Get any - sign 
	OVER C@ '$' = IF  #16 BASE ! 1 /STRING  ENDIF	\ $ sign for hex.
	OVER C@ '#' = IF  #10 BASE ! 1 /STRING  ENDIF	\ # sign for decimal
	DUP  0<= IF  -R R> BASE ! 0 EXIT  ENDIF  	\ Length 0 or less?
	0. 2SWAP
	BEGIN  
	   >NUMBER  ( ud1 c-addr1 u1 -- ud2 c-addr2 u2 )
	   DUP  IF OVER C@ '.' = IF 1 /STRING DUP DPL ! \ handle point. 
			       ELSE -R R> BASE ! 0 EXIT	
			      ENDIF   			\ Error if anything but decimal point  
	     ENDIF
	   DUP 0= 
	UNTIL 2DROP 
	R> IF  DNEGATE  ENDIF
	R> BASE ! -1 ;


-- Note #1: double numbers can't be compiled.
-- Note #2: numbers larger than 32 bits 
: CROSS-COMPILE
	ONLY  TARGET DEFINITIONS  ALSO TRANSIENT	\ Restrict search order.
	BEGIN 
	 BL WORD 
	 DUP C@ 0= 
	     IF  DROP REFILL DROP			\ Get new word, If empty, get new line.
	   ELSE  DUP COUNT S" END-CROSS" COMPARE	\ Exit cross compiler on END-CROSS
	  	 0= IF	ONLY  FORTH ALSO 
		 	FORTH DEFINITIONS		\ Normal search order again.
			DROP  REFILL DROP  EXIT
	  	 ENDIF
	  	 FIND IF EXECUTE			\ Execute if found.
	 	    ELSE NUMBER? 0= ABORT" Undefined word" 
	   		 STATE-T @ IF   LITERAL-T	\ If compiling then compile as a literal. 
					DPL @ -1 <> ABORT" cross :: 128 bit literals not supported"
				 ELSE   ( -- x 0|-1 ) DROP ( always put a 32 bit number on the stack )
	   			ENDIF
	  	   ENDIF
	  ENDIF
	AGAIN ;

-- PART 9: CROSS COMPILING THE KERNEL 

-- Not a single byte of the new Forth kernel has yet been compiled. 

ALSO TRANSIENT DEFINITIONS
  FORWARD COLD
  FORWARD WARM
PREVIOUS DEFINITIONS

S" kernel.frt" INCLUDED

-- PART 10: FINISHING AND SAVING THE TARGET IMAGE.

-- Resolve the forward references created by the cross compiler.
RESOLVE (DO) 
RESOLVE DOVAR
RESOLVE (LOOP) 
RESOLVE (.") 
RESOLVE COLD   
RESOLVE WARM
RESOLVE (ABORT") 
RESOLVE (POSTPONE)

-- Store appropriate values into some of the new Forth's variables.
TLINK @ U>D T' FORTH-WORDLIST >BODY-T !-T 
THERE   U>D T' DP             >BODY-T !-T 

SAVE-IMAGE kernel.img	-- Save the newly constructed Forth system to disk.
 
		DEPRIVE

			        ( End of File )
