/* SOD64 the Stack Oriented Design computer.
   Copyright 1994 L.C. Benschop, Eindhoven, The Netherlands.
   The program is released under the GNU General Public License version 2.
   There is NO WARRANTY.
*/

#include <unistd.h>
#include "sod64.h"

/* Perform byte swap of 64-bit words in region of memory of virtual machine*/
void swap_mem(UNS64 start,UNS64 len) {
#ifndef BIG_ENDIAN
 UNS64 i,m;
 for(i=0;i<len;i+=8) {
  m=CELL(start+i);
  m=(m>>32)|(m<<32);
  m=((m&0xffff0000ffff0000L)>>16)|((m&0x0000ffff0000ffffL)<<16);
  m=((m&0xff00ff00ff00ff00L)>> 8)|((m&0x00ff00ff00ff00ffL)<< 8);
  CELL(start+i)=m;
 } 
#endif
}

/* Multiply 64-bit unsigned numbers *a and *b.
   High half of 128-bit result in *a, low half in *b
*/   
static void umul(UNS64 *a,UNS64 *b)
{
 UNS64 ah,al,bh,bl,ph,pm1,pm2,pl;
 ah=*a>>32;al=*a&0xffffffff;
 bh=*b>>32;bl=*b&0xffffffff;
 pl=al*bl;
 if((ah|bh)==0) {
  ph=0;
 } else {
  pm1=al*bh;
  pm2=ah*bl;
  ph=ah*bh;
  pl=pl+(pm1<<32);
  ph+=(pl<(pm1<<32));
  pl=pl+(pm2<<32);
  ph+=(pl<(pm2<<32));
  ph=ph+(pm1>>32)+(pm2>>32);
 } 
 *a=ph;*b=pl;
}

/* Divide 128-bit unsigned number (high half *b, low half *c) by
   64-bit unsigned number in *a. Quotient in *b, remainder in *c.
*/
static void udiv(UNS64 *a,UNS64 *b,UNS64 *c)
{
 UNS64 d,qh,ql;
 int i,cy;
 qh=*b;ql=*c;d=*a;
 if(qh==0) {
  *b=ql/d;
  *c=ql%d;
 } else {
  for(i=0;i<64;i++) {
   cy=qh&0x8000000000000000L;
   qh<<=1;
   if(ql&0x8000000000000000L)qh++;
   ql<<=1;
   if(qh>=d||cy) {
    qh-=d;
    ql++;
   } 
   *c=qh;
   *b=ql; 
  }  
 }
}

void virtual_machine(void)
{
 register UNS64 sp,rp,ret,ip,ireg,t;
 rp=MEMSIZE;
 sp=MEMSIZE-1024;
 ip=0;
 for(;;) {
  ireg=CELL(ip);
  ip+=8;
  if(!(ireg&1)) {
    if(ireg&2) /*jumpz*/ {
                t=CELL(sp);sp+=8;  
                if(!t)ip=ireg&CELLMASK;
          } else /*call*/ {
            rp-=8;
            CELL(rp)=ip;
            ip=ireg&CELLMASK; 
          } 
  } else {
   ret=ireg&0x8000000000000000L;
   ireg=(ireg&0x7fffffffffffffffL)>>1;
   do {
    switch(ireg&31) { 
     case 0: /*nop*/ break;
     case 1: /*swap*/ t=CELL(sp);CELL(sp)=CELL(sp+8);
                      CELL(sp+8)=t;break;
     case 2: /*rot*/  t=CELL(sp+16);CELL(sp+16)=CELL(sp+8);
                      CELL(sp+8)=CELL(sp);CELL(sp)=t;break;
     case 3: /*0=*/   CELL(sp)=-!CELL(sp);break;
     case 4: /*negate*/ CELL(sp)=-CELL(sp);break;
     case 5: /* um* */ umul(&CELL(sp),&CELL(sp+8));
                       break;
     case 6: /* C@ */  CELL(sp)=BYTE(CELL(sp));break;                                                   
     case 7: /* @ */   CELL(sp)=CELL(CELL(sp)&CELLMASK);break;
     case 8: /* + */   CELL(sp+8)=CELL(sp)+CELL(sp+8);sp+=8;break;
     case 9: /* and */ CELL(sp+8)=CELL(sp)&CELL(sp+8);sp+=8;break;
     case 10:/* or  */ CELL(sp+8)=CELL(sp)|CELL(sp+8);sp+=8;break;
     case 11:/* xor */ CELL(sp+8)=CELL(sp)^CELL(sp+8);sp+=8;break;
     case 12:/* u<  */ CELL(sp+8)=-(CELL(sp+8)<CELL(sp));sp+=8;break;
     case 13:/* <   */ CELL(sp+8)=-((INT64)CELL(sp+8)<(INT64)CELL(sp));
                       sp+=8;break;
     case 14:/*lshift*/CELL(sp+8)<<=CELL(sp);sp+=8;break;
     case 15:/*rshift*/CELL(sp+8)>>=CELL(sp);sp+=8;break;
     case 16:/*um/mod*/if(CELL(sp)<=CELL(sp+8)) { /*overflow */
                         sp+=8;CELL(sp)=0xffffffffffffffffL;CELL(sp+8)=0xffffffffffffffffL;
                       }
                       else
                       {
                        udiv(&CELL(sp),&CELL(sp+8),&CELL(sp+16));
			sp+=8;
                       }break;
     case 17:/* +cy */ { UNS64 sum; if(CELL(sp)) { /* carry in */
                          sum=CELL(sp+16)+CELL(sp+8)+1;
                          sp+=8;
                          CELL(sp)=(sum<=CELL(sp));
                          CELL(sp+8)=sum;
                         } else {
                          sum=CELL(sp+16)+CELL(sp+8);
                          sp+=8;
                          CELL(sp)=(sum<CELL(sp));
                          CELL(sp+8)=sum;
                         }
                       }break; 
     case 18:/*scan1*/ if(CELL(sp)){ /* scan from right */
                         t=0;while(!(CELL(sp+8)&1) && t<64) {
                          CELL(sp+8)>>=1;t++;
                         }
                       } else { 
                         t=0;while(!(CELL(sp+8)&0x8000000000000000L) && t<64) {
                          CELL(sp+8)<<=1;t++;
                         } 
                       }sp+=8;CELL(sp)=t;break;
     case 19:/*special*/ t=CELL(sp);sp+=8;
                         save_sp=sp;save_ip=ip;save_rp=rp;
                         do_special(t);
                         sp=save_sp;ip=save_ip;rp=save_rp;
                         break;
     case 20:/* drop */ sp+=8;break;
     case 21:/* >r   */ rp-=8;CELL(rp)=CELL(sp);sp+=8;break;
     case 22:/* c!a  */ t=CELL(sp);BYTE(t)=CELL(sp+8);sp+=8;
                        CELL(sp)=t;break;
     case 23:/* !a   */ t=CELL(sp);CELL(t)=CELL((sp+8)&CELLMASK);sp+=8;
                        CELL(sp)=t;break;
     case 24:/* dup  */ sp-=8;CELL(sp)=CELL(sp+8);break;
     case 25:/* over */ sp-=8;CELL(sp)=CELL(sp+16);break;
     case 26:/* r@   */ sp-=8;CELL(sp)=CELL(rp);break;
     case 27:/* r>   */ sp-=8;CELL(sp)=CELL(rp);rp+=8;break;
     case 28:/*push0 */ sp-=8;CELL(sp)=0;break;
     case 29:/*push1 */ sp-=8;CELL(sp)=1;break;
     case 30:/*push4 */ sp-=8;CELL(sp)=8;break;                                                       
     case 31:/*lit*/ sp-=8;CELL(sp)=CELL(ip);ip+=8;break; 
    }
    ireg>>=5;
   } while(ireg);
   if(ret)/*return*/ {
            ip=(CELL(rp))&CELLMASK;
            rp+=8;
           } 
  }      
 }  
}

