(*
 * LANGUAGE    : ANS Forth 
 * PROJECT     : Forth Environments
 * DESCRIPTION : SOD64 virtual stack machine
 * CATEGORY    : Tool
 * AUTHOR      : L.C. Benschop, 1994
 * LAST CHANGE : Marcel Hendrix, Saturday, March 08, 2008, 10:34 AM.
 *)


	NEEDS -miscutil
	NEEDS -quads
	NEEDS -longhand

	REVISION -sod64  " SOD64 stack machine Version 0.01 "

	PRIVATES

: DH. ( ud -- ) 
	BASE @ >R HEX 
	<#  #16 0 DO # LOOP '$' HOLD  #> TYPE
	R> BASE ! ;

\ SOFTWARE SIMULATOR FOR SOD-64, THE STACK ORIENTED DESIGN PROCESSOR
\ Copyright 1994 L.C. Benschop Eindhoven, The Netherlands.
\ The program is released under the GNU General Public License version 2.
\ There is NO WARRANTY.

-- -------------------- PART 1: THE VIRTUAL MEMORY SPACE -----------------------

-- We need to define the virtual memory space and the words to access it.
-- These are very similar to the words that access the target space in
-- a cross compiler.

#256 #1024 *    =: mem_size
mem_size 1-     =: memmask	-- This mask confines an address to memory.
memmask 7 -     =: cellmask	-- This confines an address to aligned address.

  =CELL 2*      =: =TCELL	-- target cell size
: =TCELLS 	=TCELL * ;  ( u -- offs ) 

CREATE vmem 
	mem_size CHARS ALLOT	-- This space contains the target image.
       	vmem mem_size ERASE	-- Initialize it to zero.

-- Fetch and store characters in the virtual space.
-- All stack values are 64 bits wide.
: VC@ ( t-offs -- c ) memmask AND CHARS vmem + C@ U>D ;
: VC! ( c t-offs -- ) memmask AND CHARS vmem + NIP C! ;

-- Fetch and store cells in the virtual space.
-- SOD64 is big endian 64 bit so store explicitly big-endian.

: V@  ( t-offs -- x ) cellmask AND CHARS vmem + D@ BSWAP SWAP BSWAP ;
: V!  ( x t-offs -- ) cellmask AND CHARS vmem + >R BSWAP SWAP BSWAP R> D! ;

-- First we will load a Forth image into the virtual memory.
: LOAD-IMAGE ( "name" --- )
	/PARSE R/O BIN OPEN-FILE ABORT" can't open file" >R
	vmem mem_size R@ READ-FILE ABORT" can't read file" 
	CR DUP DEC. ." byte" ?s ."  image read" CR
	R> CLOSE-FILE ABORT" can't close file" ;

-- ------------- PART 2: Virtual machine registers and stack access ------------

0. DVALUE ir	\ Instruction register.
0   VALUE ip	\ Instruction pointer.
0   VALUE sp	\ Stack pointer.
0   VALUE rp	\ Return stack pointer.

: POP   ( --- x )  \ Pop x from data stack.
	sp V@  
	=TCELL +TO sp ;

: 2POP  ( -- y x ) \ Pop x,y from data stack.
	sp          V@  
	sp =TCELL + V@  
	2 =TCELLS +TO sp ;

: 2POPR ( -- x y ) \ Pop x,y from data stack. 
	sp =TCELL + V@  
	sp          V@  
	2 =TCELLS +TO sp ;

: PUSH  ( x --- )  \ Push x onto data stack.
	=TCELL -TO sp  
	sp V! ; 

: 2PUSH ( x y -- ) 
	2 =TCELLS -TO sp  
	sp =TCELL + V!
	sp          V! ;

: 2PUSHR ( x y -- ) 
	2 =TCELLS -TO sp  
	sp          V!
	sp =TCELL + V! ;

: RPOP  ( --- x )  \ Pop x from return stack.
	rp V@  
	=TCELL +TO rp ;

: RPUSH ( x --- )  \ Push x onto return stack.
	=TCELL -TO rp  
	rp V! ;

: .S-T  ( -- )
	CR ." stack: " 0 3 DO  sp I =TCELLS + V@ DH. SPACE  -1 +LOOP ;

-- ------------------ PART 3: The special instructions -------------------------

-- For the moment we cut some corners with OS access, defining only
-- functions 0 ... 20.

CREATE accept-buffer  #128 2+ CHARS ALLOT

0 VALUE #buf 
0 VALUE >buf

-- Emulate the cooked input of C.
: GET-KEY
	>buf #buf > IF accept-buffer #128 ACCEPT TO #buf  0 TO >buf  ENDIF
	>buf #buf = IF ^J ( Newline) ELSE  accept-buffer >buf + C@   ENDIF
	1 +TO >buf ;

: >realaddr ( t-addr -- addr )
	DROP memmask AND  CHARS vmem + ;

-- modes: 4 = R/W  2 = W/O  0 = R/O
--        5 = R/W  3 = W/O  1 = R/O  but BINary

: CREATE/OPEN-FILE ( --- )
	2POPR DROP NIP  
	POP >realaddr -ROT
	DUP 1 AND >R 
	CASE 2/
	  0 OF R/O R> IF  BIN  ENDIF OPEN-FILE    ENDOF
	  1 OF W/O R> IF  BIN  ENDIF CREATE-FILE  ENDOF
	  2 OF R/W R> IF  BIN  ENDIF OPEN-FILE    ENDOF
	ENDCASE  
	SWAP U>D PUSH S>D PUSH ;

: ROT'M ( --- a b c )
	POP DROP >R 
	2POP >realaddr NIP  SWAP R> ;

: UNROT'M ( a b c --- )
	ROT  S>D PUSH 
	SWAP S>D PUSH  
	     S>D PUSH ;

-- SOD64 uses file handles rather carelessly, i.e. it closes files it hasn't
-- even opened yet. We'll let it close anything not in use by the simulator. 
CR .( Warning: ?CLOSE-FILE : The number 5 is system specific.)
: ?CLOSE-FILE ( handle -- ior )
	DUP 5 > IF CLOSE-FILE 
	      ELSE DROP 0 
	     ENDIF ;

: DO_OS ( code --- )
	CASE
	  0 OF  CR ." exiting simulator" QUIT 	 		ENDOF
	  1 OF  GET-KEY 			S>D PUSH	ENDOF
	  2 OF  POP DROP EMIT	 				ENDOF
	  7 OF  CREATE/OPEN-FILE 		  		ENDOF
	  8 OF  POP DROP ?CLOSE-FILE  		S>D PUSH	ENDOF
	  9 OF  ROT'M WRITE-LINE      		S>D PUSH	ENDOF
	#10 OF  ROT'M READ-LINE  UNROT'M	  		ENDOF
	#11 OF  ROT'M WRITE-FILE      		S>D PUSH	ENDOF
	#12 OF  ROT'M READ-FILE  SWAP S>D PUSH  S>D PUSH 	ENDOF
	#13 OF  2POPR DROP NIP DELETE-FILE      S>D PUSH 	ENDOF
	#14 OF  POP DROP >R  2POPR DROP NIP R>  
	 	    REPOSITION-FILE 	        S>D PUSH 	ENDOF
	#15 OF  POP DROP FILE-POSITION UNROT'M 	  		ENDOF
	#16 OF  2POP >realaddr NIP SWAP  SYSTEM  
	 		RETURNCODE @ 		S>D PUSH  ENDOF
	#17 OF  POP DROP FILE-SIZE     UNROT'M  		ENDOF
	#18 OF  TIMER-RESET			  		ENDOF
	#19 OF  .ELAPSED			  		ENDOF
	#20 OF  PAGE				  		ENDOF
		CR ip H. 2 SPACES ir DH. 2 SPACES 
		CR ." code = " DEC. TRUE ABORT" unimplemented OS function" 
	ENDCASE ;

: DO_SPECIAL ( --- )
	CASE POP DROP 
	   0 OF  sp U>D PUSH     EXIT ENDOF
	   1 OF  POP DROP TO sp  EXIT ENDOF
	   2 OF  rp U>D PUSH     EXIT ENDOF
	   3 OF  POP DROP TO rp  EXIT ENDOF
	 #32 OF  POP DROP DO_OS  EXIT ENDOF
	         CR ip H. 2 SPACES ir DH. 2 SPACES 
		 ." special = " DEC. TRUE ABORT" unimplemented special instruction" 
	ENDCASE ;

-- --------------------- PART 4: The normal instructions -----------------------

: DO_NOOP ;
: DO_0=		POP  D0= S>D   PUSH ;
: DO_NEGATE	POP  DNEGATE   PUSH ;
: DO_+		2POP D+        PUSH ;
: DO_AND 	2POP DAND      PUSH ;
: DO_OR		2POP DOR       PUSH ;
: DO_XOR	2POP DXOR      PUSH ;
: DO_C@		POP  DROP VC@  PUSH ;
: DO_@		POP  DROP V@   PUSH ;
: DO_SWAP	2POPR 2PUSH ;
: DO_ROT	2POP POP 2>R  2PUSH 2R> PUSH ;
: DO_UM*	2POP UDM* 2PUSHR ;
: DO_U<		2POPR DU< S>D PUSH ;
: DO_<		2POPR D<  S>D PUSH ;
: DO_LSHIFT	2POPR DROP DLSHIFT  PUSH ;
: DO_RSHIFT	2POPR DROP DRSHIFT  PUSH ;
: DO_UM/MOD	POP 2POPR 2ROT  UMD/MOD     	     2PUSHR ;
: DO_+CY	POP DROP 1 AND >R  2POP R> UD+c U>D  2PUSHR ;
: DO_SCAN1 	TRUE ABORT" SCAN1 unimplemented " ;
: DO_DROP	POP 2DROP ;
: DO_>R		POP RPUSH ;
: DO_C!A	2POPR 2DUP 2>R  DROP VC!  2R> PUSH ;
: DO_!A		2POPR 2DUP 2>R  DROP V!   2R> PUSH ;
: DO_DUP	POP   2DUP 2PUSH ;
: DO_OVER	2POP  2DUP 2>R 2PUSH 2R> PUSH ;
: DO_R@		RPOP  2DUP RPUSH PUSH ;
: DO_R>		RPOP PUSH ;
: DO_0		0. PUSH ;
: DO_1		1. PUSH ;
: DO_8		8. PUSH ;
: DO_LIT	ip V@ PUSH  =TCELL +TO ip ;

CREATE opcodes
 ' DO_NOOP , 	' DO_SWAP , 	' DO_ROT , 	' DO_0= ,
 ' DO_NEGATE , 	' DO_UM* , 	' DO_C@ , 	' DO_@ ,
 ' DO_+ ,  	' DO_AND , 	' DO_OR , 	' DO_XOR ,
 ' DO_U< , 	' DO_< ,  	' DO_LSHIFT , 	' DO_RSHIFT ,
 ' DO_UM/MOD , 	' DO_+CY , 	' DO_SCAN1 , 	' DO_SPECIAL ,
 ' DO_DROP , 	' DO_>R , 	' DO_C!A ,	' DO_!A ,
 ' DO_DUP , 	' DO_OVER , 	' DO_R@ , 	' DO_R> ,
 ' DO_0 , 	' DO_1 , 	' DO_8 , 	' DO_LIT ,


-- --------------------- PART 4a: The SEE instructions --------------------

: SDO_NOOP 	;
: SDO_SWAP	." SWAP " ;
: SDO_ROT	." ROT " ;
: SDO_0=	." 0= " ;
: SDO_NEGATE	." NEGATE " ;
: SDO_UM*	." UM* " ;
: SDO_C@	." C@ " ;
: SDO_@		." @ " ;
: SDO_+		." + " ;
: SDO_AND 	." AND " ;
: SDO_OR	." OR " ;
: SDO_XOR	." XOR " ;
: SDO_U<	." U< " ;
: SDO_<		." < " ;
: SDO_LSHIFT	." LSHIFT " ;
: SDO_RSHIFT	." RSHIFT " ;
: SDO_UM/MOD	." UM/MOD " ;
: SDO_+CY	." +CY " ;
: SDO_SCAN1 	." SCAN1 " ;
: SDO_DROP	." DROP " ;
: SDO_>R	." >R " ;
: SDO_C!A	." C!A " ;
: SDO_!A	." !A " ;
: SDO_DUP	." DUP " ;
: SDO_OVER	." OVER " ;
: SDO_R@	." R@ " ;
: SDO_R>	." R> " ;
: SDO_0		." 0 " ;
: SDO_1		." 1 " ;
: SDO_8		." 8 " ;
: SDO_LIT	ip V@  =TCELL +TO ip  D. ;
: SDO_SPECIAL 	." SPECIAL " ;

CREATE seers
 ' SDO_NOOP , 	' SDO_SWAP , 	' SDO_ROT , 	' SDO_0= ,
 ' SDO_NEGATE , ' SDO_UM* , 	' SDO_C@ , 	' SDO_@ ,
 ' SDO_+ ,  	' SDO_AND , 	' SDO_OR , 	' SDO_XOR ,
 ' SDO_U< , 	' SDO_< ,  	' SDO_LSHIFT , 	' SDO_RSHIFT ,
 ' SDO_UM/MOD , ' SDO_+CY , 	' SDO_SCAN1 , 	' SDO_SPECIAL ,
 ' SDO_DROP , 	' SDO_>R , 	' SDO_C!A ,	' SDO_!A ,
 ' SDO_DUP , 	' SDO_OVER , 	' SDO_R@ , 	' SDO_R> ,
 ' SDO_0 , 	' SDO_1 , 	' SDO_8 , 	' SDO_LIT ,

-- ---------------- PART 5a: The virtual SEE machine --------------------------

: DO_SEE
	CR ip H. 2 SPACES ip V@ DH. 2 SPACES
	ip V@ TO ir  =TCELL +TO ip							\ Instruction fetch.
	ir DROP 3 AND 0 = IF ." CALL " ir DROP 		    H. EXIT  ENDIF 		\ Call
	ir DROP 3 AND 2 = IF ." JZ   " ir DROP cellmask AND H. EXIT  ENDIF		\ Conditional jump
	ir 1 DRSHIFT TO ir
	#12 0 DO
		 ir DROP #31 AND seers []CELL PERFORM 
		 ir 5 DRSHIFT TO ir
	    LOOP
	ir DROP 4 AND IF  ." ; "  ENDIF ;						\ Return

: IDIS ( t-offs -- )
	TO ip
	BEGIN  DO_SEE  EKEY ESC = UNTIL ;
	
: DUMP ( t-offs count -- ) 
	=TCELLS BOUNDS 
	         ?DO  CR I H. 2 SPACES  
		         I V@ DH. 2 SPACES 
		         I 8 BOUNDS ?DO  I VC@ DROP SEMIT  LOOP 
	=TCELL +LOOP ;

-- ---------------- PART 5: The virtual machine itself ------------------------

: DO_INSTR
	ip V@ TO ir  =TCELL +TO ip							\ Instruction fetch.
	ir DROP 3 AND 0 = IF ip U>D RPUSH  ir DROP TO ip  EXIT	ENDIF 			\ Call
	ir DROP 3 AND 2 = IF POP D0= IF  ir DROP cellmask AND TO ip  ENDIF  EXIT  ENDIF	\ Conditional jump
	ir 1 DRSHIFT TO ir
	#12 0 DO
		 ir DROP #31 AND opcodes []CELL PERFORM
		 ir 5 DRSHIFT TO ir
	    LOOP
	ir DROP 4 AND IF  RPOP DROP TO ip  ENDIF ;					\ Return

-- ------------------------- PART 6: Running it all ---------------------------

: SOD64 ( "kernel-image-name" --- )
	LOAD-IMAGE
	CLEAR ip
	mem_size 	 TO rp
	mem_size #1024 - TO sp
	BEGIN  DO_INSTR  AGAIN ;

:ABOUT 	CR ." try: SOD64      kernel.img "
	CR ."  or: LOAD-IMAGE kernel.img "
	CR ."  or  ( u -- )      IDIS"
	CR ."  or  ( addr u -- ) DUMP" ;

		.ABOUT -sod64 CR 
		DEPRIVE

			        ( End of File )
